#ifndef INCLUDED_SDSL_HYBRID_VECTOR
#define INCLUDED_SDSL_HYBRID_VECTOR

#include "int_vector.hpp"
#include "util.hpp"
#include "iterators.hpp"

#include <cstdlib>
#include <vector>
#include <algorithm>
#include <iostream>

namespace sdsl
{
template<uint8_t t_b=1, uint16_t sb_rate=16> class rank_support_hybrid;
template<uint8_t t_b=1, uint16_t sb_rate=16> class select_support_hybrid;

template<uint16_t sb_rate=16> class hybrid_vector {
public:
  typedef bit_vector::size_type size_type;
  typedef bit_vector::value_type value_type;
  typedef bit_vector::difference_type difference_type;
  typedef random_access_const_iterator<hybrid_vector> iterator;
  typedef rank_support_hybrid<1, sb_rate>   rank_1_type;
  typedef rank_support_hybrid<0, sb_rate>   rank_0_type;
  typedef select_support_hybrid<1, sb_rate> select_1_type;
  typedef select_support_hybrid<0, sb_rate> select_0_type;
  friend class rank_support_hybrid<1, sb_rate>;
  friend class rank_support_hybrid<0, sb_rate>;
  friend class select_support_hybrid<1, sb_rate>;
  friend class select_support_hybrid<0, sb_rate>;

  static const int k_block_size = 255;
  static const long k_block_bytes = 32L;
  static const int k_sb_rate = sb_rate;
  static const int k_header_size = 8 + 2 * k_sb_rate;
  static const int k_sblock_size = (k_block_size * k_sb_rate);

  hybrid_vector(const bit_vector& bv) {
    m_length = (int)bv.size();
    n_block = (m_length + k_block_size - 1) / k_block_size;
    n_sblock = (n_block + k_sb_rate - 1) / k_sb_rate;
    m_header = new unsigned char[(n_sblock + 1) * k_header_size];
    std::fill(m_header, m_header + (n_sblock + 1) * k_header_size, 0);
    compute_trunk_size(bv);
    m_trunk = new unsigned char[trunk_size];

    long sb_count = 0;
    unsigned char *buf = new unsigned char[k_block_size];
    for (long beg = 0, tot_rank = 0, block_id = 0, trunk_ptr = 0; beg < m_length; beg += k_block_size, ++block_id) {
      long end = std::min(beg + k_block_size, m_length);
      int sb_id = block_id / k_sb_rate;

      // Update super block information.
      if ((block_id % k_sb_rate) == 0) {
        *((int *)(m_header + k_header_size * sb_id)) = (int)trunk_ptr;
        *((int *)(m_header + k_header_size * sb_id + 4)) = (int)tot_rank;
        if (sb_id > 0 && (sb_count == 0 || sb_count == k_sblock_size))
           *((int *)(m_header + k_header_size * (sb_id - 1))) |= 0x80000000; // uniform sb
        sb_count = 0;
      }

      // Copy bytes [beg..end) to buffer (pad with zeros if necessary).
      for (long j = beg; j < end; ++j) buf[j - beg] = '0' + bv[j];
      std::fill(buf + end - beg, buf + k_block_size, '0');

      // Count 1s and runs.
      int ones = 0, runs = 1;
      for (long i = 0; i < k_block_size; ++i) {
        if (buf[i] == '1') ++ones;
        if (i + 1 != k_block_size && buf[i + 1] != buf[i]) ++runs;
      }

      // Store block popcount.
      m_header[sb_id * k_header_size + 8 + ((block_id % k_sb_rate) * 2) + 1] = ones;
      if (ones == 255) m_header[sb_id * k_header_size + 8 + ((block_id % k_sb_rate) * 2)] |= 0x80;
      if (0 < ones && ones < 255) { // non uniform block
        int advanced_encoding_size = std::min(std::min(ones, k_block_size - ones), std::max(0, runs - 2));
        if (k_block_bytes <= advanced_encoding_size) { // plain encoding
          m_header[sb_id * k_header_size + 8 + ((block_id % k_sb_rate) * 2)] |= k_block_bytes;
          for (int i = 0; i < k_block_size; i += 64) {
            unsigned long word = 0UL;
            for (int j = 0; j < 64; ++j)
              if (i + j < k_block_size && buf[i + j] == '1') word |= (1UL << (63 - j));
              *((unsigned long* )(m_trunk + trunk_ptr)) = word;
              trunk_ptr += 8;
          }
        } else {
          if (std::max(0, runs - 2) < std::min(ones, k_block_size - ones)) { // runs encoding
            int nbytes_runs = std::max(0, runs - 2);
            m_header[sb_id * k_header_size + 8 + ((block_id % k_sb_rate) * 2)] |= nbytes_runs;
            m_header[sb_id * k_header_size + 8 + ((block_id % k_sb_rate) * 2)] |= (buf[0] == '1') ? 0x80 : 0x00;

            for (int i = 0, t = 0; t < nbytes_runs; ++i)
              if (buf[i + 1] != buf[i]) { m_trunk[trunk_ptr++] = (unsigned)i; ++t; }
          } else { // minority encoding
            m_header[sb_id * k_header_size + 8 + ((block_id % k_sb_rate) * 2)] |= std::min(ones, k_block_size - ones);
            m_header[sb_id * k_header_size + 8 + ((block_id % k_sb_rate) * 2)] |= 0x40;
            if (ones < k_block_size - ones)
              m_header[sb_id * k_header_size + 8 + ((block_id % k_sb_rate) * 2)] |= 0x80;
            unsigned char key_char = (ones < k_block_size - ones) ? '1' : '0';
            for (int i = 0; i < k_block_size; ++i) if (buf[i] == key_char) m_trunk[trunk_ptr++] = i;
          }
        }
      }

      // Update global rank.
      tot_rank += ones;
      sb_count += ones;
    }

    delete[] buf; 
  }
  
  value_type operator[](size_type i) const { return 0; }
  hybrid_vector& operator = (const hybrid_vector& hybrid) { if (this != &hybrid) copy(hybrid); return *this; }
  hybrid_vector& operator = (hybrid_vector&& hybrid) { swap(hybrid); return *this; }
  size_type size() const { return m_length; }  
  void load(std::istream& in) { m_length = 0; }
  iterator begin() const { return iterator(this, 0); }
  iterator end() const { return iterator(this, size()); }
  hybrid_vector() { m_length = 0; }
  hybrid_vector(const hybrid_vector& hybrid) { copy(hybrid); }
  hybrid_vector(hybrid_vector&& hybrid) { exit(1); }
  void copy(const hybrid_vector& hybrid) { exit(1); }
  ~hybrid_vector() { if (m_length) { delete[] m_trunk; delete[] m_header; } }

  void swap(hybrid_vector& hybrid) {
    if (this == &hybrid) return;
    std::swap(m_trunk, hybrid.m_trunk);
    std::swap(m_header, hybrid.m_header);
    std::swap(m_length, hybrid.m_length);
    std::swap(n_block, hybrid.n_block);
    std::swap(n_sblock, hybrid.n_sblock);
    std::swap(trunk_size, hybrid.trunk_size);
  }
  
  size_type serialize(std::ostream& out, structure_tree_node* v=nullptr, std::string name="") const {
    return (trunk_size + k_header_size * n_sblock);
  }

private:
  void compute_trunk_size(const bit_vector& bv) {
    unsigned char *buf = new unsigned char[k_block_size];
    trunk_size = 0;
    for (long beg = 0; beg < m_length; beg += k_block_size) {
      long end = std::min(beg + k_block_size, m_length);
      for (int j = beg; j < end; ++j) buf[j - beg] = '0' + bv[j];
      std::fill(buf + end - beg, buf + k_block_size, '0');
      long ones = 0, runs = 1;
      for (long i = 0; i < k_block_size; ++i) {
        if (buf[i] == '1') ++ones;
        if (i + 1 != k_block_size && buf[i + 1] != buf[i]) ++runs;
      }
      long advanced_encoding_size = std::min(std::min(ones, k_block_size - ones), std::max(0L, runs - 2));
      trunk_size += std::min(k_block_bytes, advanced_encoding_size);
    }
    delete[] buf;
  }

  unsigned char *m_trunk, *m_header;
  long m_length;                     // bitvector length (in bits)
  long n_block, n_sblock;            // number of blocks
  long trunk_size;                   // size of proper encoding
};

/////////////////////////////// RANK ///////////////////////////////////////////

template<uint8_t t_b, uint16_t sb_rate> class rank_support_hybrid {
public:
  static const int k_block_size = 255;
  static const long k_block_bytes = 32L;  
  static const int k_sb_rate = sb_rate;
  static const int k_header_size = 8 + 2 * k_sb_rate;
  static const int k_sblock_size = (k_block_size * k_sb_rate);

  typedef hybrid_vector<sb_rate> bit_vector_type;
  typedef typename bit_vector_type::size_type size_type;
  explicit rank_support_hybrid(const bit_vector_type* v=nullptr) { set_vector(v); }

  const size_type rank(size_type i)const {
    if (i <= 0) return 0;

    // Read super-block header.
    long block_idx = (i - 1) / k_block_size;
    int local_i = i - block_idx * k_block_size;
    long sb_idx = block_idx / sb_rate;
    unsigned char *header_ptr = m_header + (sb_idx * k_header_size);
    int *header_int_ptr = (int *)header_ptr;
    int trunk_ptr = (*header_int_ptr) & 0x3fffffff;
    int sb_rank = *(header_int_ptr + 1);
    header_ptr += 8;

    // Uniform super-block optimization.
    if ((*header_int_ptr) & 0x80000000) 
      return sb_rank + ((*(header_ptr + 1)) & 0x01) * (i - sb_idx * k_sblock_size);

    // Fast-forward through preceding blocks in the super-block.
    long block_rank = 0;
    for (long blk = sb_idx * k_sb_rate; blk != block_idx; ++blk) {
      trunk_ptr += (*header_ptr) & 0x3F; // Increase trunk pointer.
      ++header_ptr;
      block_rank += *header_ptr; // Accumulate 1s in the block.
      ++header_ptr;
    }

    unsigned char *trunk_p = m_trunk + trunk_ptr;
    unsigned char header_byte = (*header_ptr);
    int encoding_size = header_byte & 0x3F;
    ++header_ptr;
    int ones = (*header_ptr);

    // #runs <= 2
    int first_bit = ((header_byte & 0x80) >> 7) & 0x01;
    if (!encoding_size) {
      int how_many_bits = first_bit * ones + (1 - first_bit) * (k_block_size - ones);
      int local_rank = std::min(local_i, how_many_bits);
      return sb_rank + block_rank + (first_bit * local_rank + (1 - first_bit) * (local_i - local_rank));
    }
    
    // #runs > 2
    if (encoding_size < k_block_bytes) {
      if (header_byte & 0x40) { // minority encoding
        int tot = 0;
        while (tot < encoding_size && (*trunk_p++) < local_i) ++tot;
        return sb_rank + block_rank + first_bit * tot + (1 - first_bit) * (local_i - tot);
      }
      // Runs encoding
      if (header_byte & 0x80) {
        int j = 0, last = -1, acc = 0;
        while (j + 1 < encoding_size && *(trunk_p + 1) < local_i) {
          acc += *trunk_p - last; ++trunk_p;
          last = *trunk_p; ++trunk_p; j += 2;
        }
        if (j + 1 >= encoding_size) {
          if (j < encoding_size) {
            if (*trunk_p  >= local_i) acc += local_i - last - 1;
            else {
              acc += (*trunk_p) - last;
              acc += (ones - acc) - std::min(ones - acc, k_block_size - local_i);
            }
          } else acc += std::min(ones - acc, local_i - last - 1);
        } else acc += std::min((int)(*trunk_p), local_i - 1) - last;
        return sb_rank + block_rank + acc;
      } else {
        int j = 0, last = -1, acc = 0;
        while (j + 1 < encoding_size && *(trunk_p + 1) < local_i) {
          acc += *trunk_p - last; ++trunk_p;
          last = *trunk_p; ++trunk_p; j += 2;
        }
        if (j + 1 >= encoding_size) {
          if (j < encoding_size) {
            if (*trunk_p  >= local_i) acc += local_i - last - 1;
            else {
              acc += (*trunk_p) - last;
              acc += ((k_block_size - ones) - acc) - std::min((k_block_size - ones) - acc, k_block_size - local_i);
            }
          } else acc += std::min((k_block_size - ones) - acc, local_i - last - 1);
        } else acc += std::min((int)(*trunk_p), local_i - 1) - last;
        return sb_rank + block_rank + (local_i - acc);
      }
    } else { // plain decoding
      unsigned long *trunk_l_ptr = (unsigned long *)(m_trunk + trunk_ptr);
      int j = 0, bit = 0;
      for (; j + 8 < k_block_bytes && bit + 64 <= local_i; bit += 64, j += 8)
        block_rank += __builtin_popcountll(*trunk_l_ptr++);
      if (bit != local_i) block_rank += __builtin_popcountll((*trunk_l_ptr) >> (64 - local_i + bit));
      return sb_rank + block_rank;
    }
  }

  const size_type operator()(size_type i) const { return rank(i); }
  const size_type size() const { return m_length; }
  void load(std::istream&, const bit_vector_type* v=nullptr) { set_vector(v); }
  size_type serialize(std::ostream&, structure_tree_node* v=nullptr, std::string name="") const {return 0;}

  void set_vector(const bit_vector_type* v=nullptr) {
    if (v==nullptr) return;
    m_trunk = v->m_trunk;
    m_header = v->m_header;
    m_length = v->m_length;
    n_block = v->n_block;
    n_sblock = v->n_sblock;
    trunk_size = v->trunk_size;
  }
  
  rank_support_hybrid& operator=(const rank_support_hybrid& rs) {
    if (this != &rs) set_vector(rs.m_v);
    return *this;
  }

  void swap(rank_support_hybrid& rs) {
    std::swap(m_trunk, rs.m_trunk);
    std::swap(m_header, rs.m_header);
    std::swap(m_length, rs.m_length);
    std::swap(n_block, rs.n_block);
    std::swap(n_sblock, rs.n_sblock);
    std::swap(trunk_size, rs.trunk_size);
  }

private:
  unsigned char *m_trunk, *m_header;
  long m_length;                     // bitvector length (in bits)
  long n_block, n_sblock;            // number of blocks
  long trunk_size;                   // size of proper encoding
};

///////////////////////////////// SELECT ///////////////////////////////////////

template<uint8_t t_b, uint16_t sb_rate> class select_support_hybrid {
public:
  typedef hybrid_vector<sb_rate> bit_vector_type;
  typedef typename bit_vector_type::size_type size_type;
  explicit select_support_hybrid(const bit_vector_type* v = nullptr) { }
  size_type select(size_type i) const { return 0; }
  const size_type operator()(size_type i) const { return 0; }
  const size_type size() const { return 0; }
  void set_vector(const bit_vector_type* v=nullptr) { }
  select_support_hybrid& operator=(const select_support_hybrid& rs) { return *this; }
  void swap(select_support_hybrid&) { }
  void load(std::istream&, const bit_vector_type* v=nullptr) { set_vector(v); }
  size_type serialize(std::ostream&, structure_tree_node* v=nullptr, std::string name="") const { return 0; }
};

} // end namespace sdsl

#endif // INCLUDED_SDSL_HYBRID_VECTOR
