#include <cstdlib>
#include <iostream>
#include <fstream>
#include <cstdio>
#include <unistd.h>
#include <cstring>
#include <algorithm>

#include <sdsl/bit_vectors.hpp>
#include <sdsl/csa_wt.hpp>
#include <sdsl/wt_huff.hpp>
#include <sdsl/suffix_arrays.hpp>
#include <sdsl/hybrid_vector.hpp>

#include "utils.h"

// Test count queries using given index.
template<typename index_type>
void test(index_type &index,
    std::string index_name,
    long double length_MiB,   // length of input sequence (in MiB)
    int n_query,              // number of queries
    int pat_length,           // query pattern length
    unsigned char *query_data // queries to be used for tests
  ) {
  fprintf(stderr, "%20s", index_name.c_str());
  
  long double size_in_MiB = sdsl::size_in_mega_bytes(index);
  fprintf(stderr, "%9.2LfMiB", size_in_MiB); // print abs size
  fprintf(stderr, "%11.2Lf%%", (100.L * size_in_MiB) / length_MiB); // rel size
  
  long checksum = 0;
  long double query_start = utils::wclock();

  static const int reps = 30; // we do 30 repetitions to get good average time

  for (int i = 0, ii = 0; i < n_query * reps; ++i, ++ii) {

    unsigned char *pat_beg = query_data + (ii * pat_length);
    unsigned char *pat_end = pat_beg + pat_length;

    if (ii == n_query) ii = 0;
 
    int count = sdsl::count(index, pat_beg, pat_end);
    checksum += count; // prevents compiler optimizations, does not affect time
  }

  long double queries_time = utils::wclock() - query_start;
  long double symbol_time = (long double)queries_time / (n_query * pat_length);
  symbol_time = (symbol_time * 1000000.L) / reps;

  fprintf(stderr, "%11.2Lfs", queries_time);  // total wallclock time for testing
  fprintf(stderr, "%11.4Lfs\n", symbol_time); // normalized time to match a single
                                              // symbol during count query
  
  // To prevent compiler optimization.
  if (checksum == 89032890438920L)
    std::exit(EXIT_FAILURE);
}

int main(int argc, char **argv) {
  srand(time(0) + getpid());
  fprintf(stderr,  "Count queries benchmark\n\n");

  const int n_query = 50000; // we extract 50K random patterns from text
  const int pat_length = 20; // patterns are all of length 20

  std::ifstream f("../files/list");
  while (!f.eof()) {
    char tmp[100];
    f >> tmp;

    std::string fname = "../files/" + std::string(tmp);
    fprintf(stderr, "%s", fname.c_str());

    // Read text.
    long length;
    unsigned char *text;
    utils::read_file(text, length, fname.c_str());
    long double length_MiB = length / (1024.L * 1024);

    fprintf(stderr, " (size = %.2LfMiB)\n", length_MiB);
    fprintf(stderr, "%20s%12s%12s%12s%12s\n",
      "name", "abs-size", "rel-size", "abs-time", "query-time");

    // Generate queries
    unsigned char *query_data = new unsigned char[n_query * pat_length];
    for (int i = 0, ptr = 0; i < n_query; ++i, ptr += pat_length) {
      int start = utils::random_int(0, length - pat_length);
      for (int j = 0; j < pat_length; ++j)
        query_data[ptr + j] = text[start + j];
    }

    /* RRR tests *************************************************************/
    {
      sdsl::csa_wt<
        sdsl::wt_huff< sdsl::rrr_vector<15> >, // Huffman-shaped wavelet tree
                                               // for balanced WT use wt_blcd
        1000000, // SA sampling rate, irrelevant here, any huge value will do
        1000000  // ISA sampling rate
      > csa;
      construct(csa, fname, 1);    
      test(csa, "rrr_vector<15>", length_MiB, n_query, pat_length, query_data);
    }

    {
      sdsl::csa_wt<
        sdsl::wt_huff<sdsl::rrr_vector<31> >,
        1000000,
        1000000
      > csa;
      construct(csa, fname, 1);    
      test(csa, "rrr_vector<31>", length_MiB, n_query, pat_length, query_data);
    }

    {
      sdsl::csa_wt<
        sdsl::wt_huff<sdsl::rrr_vector<63> >,
        1000000,
        1000000
      > csa;
      construct(csa, fname, 1);    
      test(csa, "rrr_vector<63>", length_MiB, n_query, pat_length, query_data);
    }

    {
      sdsl::csa_wt<
        sdsl::wt_huff<sdsl::rrr_vector<127> >,
        1000000,
        1000000
      > csa;
      construct(csa, fname, 1);    
      test(csa, "rrr_vector<127>", length_MiB, n_query, pat_length, query_data);
    }
    /* End of RRR tests ******************************************************/
    
    /* Hybrid bitvector tests ************************************************/
    {
      sdsl::csa_wt<
        sdsl::wt_huff<sdsl::hybrid_vector<8> >,
        1000000,
        1000000
      > csa;
      construct(csa, fname, 1);    
      test(csa, "hybrid_vector<8>", length_MiB, n_query, pat_length, query_data);
    }

    {
      sdsl::csa_wt<
        sdsl::wt_huff<sdsl::hybrid_vector<16> >,
        1000000,
        1000000
      > csa;
      construct(csa, fname, 1);    
      test(csa, "hybrid_vector<16>", length_MiB, n_query, pat_length, query_data);
    }

    {
      sdsl::csa_wt<
        sdsl::wt_huff<sdsl::hybrid_vector<32> >,
        1000000,
        1000000
      > csa;
      construct(csa, fname, 1);    
      test(csa, "hybrid_vector<32>", length_MiB, n_query, pat_length, query_data);
    }

    {
      sdsl::csa_wt<
        sdsl::wt_huff<sdsl::hybrid_vector<64> >,
        1000000,
        1000000
      > csa;
      construct(csa, fname, 1);    
      test(csa, "hybrid_vector<64>", length_MiB, n_query, pat_length, query_data);
    }
    /* End of hybrid bitvector tests *****************************************/
    
    // Uncompressed bitvector test ////////////////////////////////////////////
    {
      sdsl::csa_wt<
        sdsl::wt_huff<
          sdsl::bit_vector,
          sdsl::rank_support_v5<>,
          sdsl::select_support_scan<1, 1>,
          sdsl::select_support_scan<0, 1>
        >,
        1000000,
        1000000
      > csa;
      construct(csa, fname, 1);    
      test(csa, "uncompressed", length_MiB, n_query, pat_length, query_data);
    }

    // RLE-compressed wavelet tree test ///////////////////////////////////////
    {
      sdsl::csa_wt<
        sdsl::wt_rlmn<>,
        1000000,
        1000000
      > csa;
      construct(csa, fname, 1);
      test(csa, "wt_rlmn", length_MiB, n_query, pat_length, query_data);
    }

    fprintf(stderr, "\n");
  }

  f.close(); // close file list
}

